#! python2
import Leap
import sys, thread, time
import ctypes
import os
import numpy
import csv
import math
from Leap import CircleGesture, KeyTapGesture, ScreenTapGesture, SwipeGesture

# Este archivo contiene el script utilizado para las grabaciones y reconocimiento de gestos usando el Leap Motion Controller (sin usar machine learning), a traves de la similitud del coseno
# Todos los comentarios estan tanto en espanol como en ingles

# This file contains the script used for recordings and recognition of gestures using the Leap Motion Controller (without using machine learning), through the similarity of the cosine.
# All the comments are written in spanish and english

#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

dataSet = {} # Inicializamos un diccionario donde se guardaran los gestos a comparar
			 # We initialized a dictionary where the gestures to be compared will be stored.

class LeapMotionListener(Leap.Listener):
	finger_names = ['Thumb', 'Index', 'Middle', 'Ring', 'Pinky'] 			# Lista con los nombres de los dedos
																			# List with the names of the fingers
	bone_names = ['Metacarpal', 'Proximal', 'Intermediate', 'Distal']		# Lista con los nombres de los huesos
																			# List with the names of the bones
	state_names = ['STATED_INVALID', 'STATE_START', 'STATE_UPDATE', 'STATE_END']	# Lista con los nombres de los estados
																					# List with the states
	
	global vectorCalculated, vectorCalculated2, euclideanDistance, cosine_similarity, csvTest, normalization
	

#------------------------------------ COSINE SIMILARITY ALGORITHM  BEGIN ------------------------------------
	
	def euclideanDistance(xUno, yUno, zUno, xDos, yDos, zDos): # Realiza las distancia de euclides entre la X Y Z uno (tres primeros parametros) y la X Y Z dos (tres ultimos parametros)
															   # Calculates de euclidean distance between the X Y Z uno (first three parameters) and the X Y Z dos (last three parameters)
		x = numpy.array((xUno, yUno, zUno))		# Crea los numpy arrays para realizar los calculos
												# Numpy arrays are inicializated in order to do calculations
		y = numpy.array((xDos, yDos, zDos))
		return numpy.sqrt(numpy.sum((x-y)**2))
		
	def vectorCalculated(fingerVector): # Este metodo calcula la distancia de euclides para los vectores que solo tengan una mano
										# This method calculates the euclidean distance only for one hand vectors
		# Se crea un nuevo vector donde se almacenaran las distancias de euclides entre las diferentes combinaciones de dedos
		# A new vector is inicializated to store all the euclidean distances betwwen fingers
		finalVector = []
		
		# 1 Distancia de Euclides entre la palma y el pulgar
		# 1 Euclidean distance between the palm and the thumb
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[3], fingerVector[4], fingerVector[5]))
		# 2 Distancia de Euclides entre la palma y el indice
		# 2 Euclidean distance betwwen the palm and the index
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[6], fingerVector[7], fingerVector[8]))
		# 3 Distancia de Euclides entre la palma y el corazon
		# 3 Euclidean distance between the palm and the middle finger
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[9], fingerVector[10], fingerVector[11]))
		# 4 Distancia de Euclides entre la palma y el anular
		# 4 Euclidean distance between the palm and the ring finger
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[12], fingerVector[13], fingerVector[14]))
		# 5 Distancia de Euclides entre la palma y el menique
		# 5 Euclidean distance between the palm and the pinky
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[15], fingerVector[16], fingerVector[17]))
		# 6 Distancia de Euclides entre el pulgar y el indice
		# 6 Eunclidean distance betwwen the thumb and the index
		finalVector.append(euclideanDistance(fingerVector[3], fingerVector[4], fingerVector[5], fingerVector[6], fingerVector[7], fingerVector[8]))
		# 7 Distancia de Euclides entre el indice y el corazon
		# 7 Euclidean distance between the index and the middle finger
		finalVector.append(euclideanDistance(fingerVector[6], fingerVector[7], fingerVector[8], fingerVector[9], fingerVector[10], fingerVector[11]))
		# 8 Distancia de Euclides entre el corazon y el anular
		# 8 Euclidean distance between the middle and the ring finger
		finalVector.append(euclideanDistance(fingerVector[9], fingerVector[10], fingerVector[11], fingerVector[12], fingerVector[13], fingerVector[14]))
		# 9 Distancia de Euclides entre el anular y el menique
		# 9 Euclidean distance between the ring and the pinky finger
		finalVector.append(euclideanDistance(fingerVector[12], fingerVector[13], fingerVector[14], fingerVector[15], fingerVector[16], fingerVector[17]))
		# 10 Distancia de Euclides entre el pulgar y el menique
		# 10 Euclidean distance between the thumb and the pinky
		finalVector.append(euclideanDistance(fingerVector[3], fingerVector[4], fingerVector[5], fingerVector[15], fingerVector[16], fingerVector[17]))
		
		return finalVector
		
	def vectorCalculated2(fingerVector): # Este metido calcula la ditancia de euclides para los vectoresque tengan dos manos
										 # This method calculates the euclidean distance only for two hand vectors
		# Se crea un nuevo vecotr donde se almacenaran los distancias de euclides entre las diferentes combinaciones de dedos
		# A new vector is inicializated to store all the euclidean distances betwwen fingers
		finalVector = []
		
		# MANO 1 (izquierda)
		# HAND 1 (left)

		# 1 Distancia de Euclides entre la palma y el pulgar
		# 1 Euclidean distance between the palm and the thumb
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[3], fingerVector[4], fingerVector[5]))
		# 2 Distancia de Euclides entre la palma y el indice
		# 2 Euclidean distance betwwen the palm and the index
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[6], fingerVector[7], fingerVector[8]))
		# 3 Distancia de Euclides entre la palma y el corazon
		# 3 Euclidean distance between the palm and the middle finger
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[9], fingerVector[10], fingerVector[11]))
		# 4 Distancia de Euclides entre la palma y el anular
		# 4 Euclidean distance between the palm and the ring finger
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[12], fingerVector[13], fingerVector[14]))
		# 5 Distancia de Euclides entre la palma y el menique
		# 5 Euclidean distance between the palm and the pinky
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[15], fingerVector[16], fingerVector[17]))
		# 6 Distancia de Euclides entre el pulgar y el indice
		# 6 Eunclidean distance betwwen the thumb and the index
		finalVector.append(euclideanDistance(fingerVector[3], fingerVector[4], fingerVector[5], fingerVector[6], fingerVector[7], fingerVector[8]))
		# 7 Distancia de Euclides entre el indice y el corazon
		# 7 Euclidean distance between the index and the middle finger
		finalVector.append(euclideanDistance(fingerVector[6], fingerVector[7], fingerVector[8], fingerVector[9], fingerVector[10], fingerVector[11]))
		# 8 Distancia de Euclides entre el corazon y el anular
		# 8 Euclidean distance between the middle and the ring finger
		finalVector.append(euclideanDistance(fingerVector[9], fingerVector[10], fingerVector[11], fingerVector[12], fingerVector[13], fingerVector[14]))
		# 9 Distancia de Euclides entre el anular y el menique
		# 9 Euclidean distance between the ring and the pinky finger
		finalVector.append(euclideanDistance(fingerVector[12], fingerVector[13], fingerVector[14], fingerVector[15], fingerVector[16], fingerVector[17]))
		# 10 Distancia de Euclides entre el pulgar y el menique
		# 10 Euclidean distance between the thumb and the pinky
		finalVector.append(euclideanDistance(fingerVector[3], fingerVector[4], fingerVector[5], fingerVector[15], fingerVector[16], fingerVector[17]))
		
		# MANO 2 (derecha)
		# HAND 2 (right)
		# 11 Distancia de Euclides entre la palma y el pulgar
		# 11 Euclidean distance between the palm and the thumb
		finalVector.append(euclideanDistance(fingerVector[18], fingerVector[19], fingerVector[20], fingerVector[21], fingerVector[22], fingerVector[23]))
		# 12 Distancia de Euclides entre la palma y el indice
		# 12 Euclidean distance betwwen the palm and the index
		finalVector.append(euclideanDistance(fingerVector[18], fingerVector[19], fingerVector[20], fingerVector[24], fingerVector[25], fingerVector[26]))
		# 13 Distancia de Euclides entre la palma y el corazon
		# 13 Euclidean distance between the palm and the middle finger
		finalVector.append(euclideanDistance(fingerVector[18], fingerVector[19], fingerVector[20], fingerVector[27], fingerVector[28], fingerVector[29]))
		# 14 Distancia de Euclides entre la palma y el anular
		# 14 Euclidean distance between the palm and the ring finger
		finalVector.append(euclideanDistance(fingerVector[18], fingerVector[19], fingerVector[20], fingerVector[30], fingerVector[31], fingerVector[32]))
		# 15 Distancia de Euclides entre la palma y el menique
		# 15 Euclidean distance between the palm and the pinky
		finalVector.append(euclideanDistance(fingerVector[18], fingerVector[19], fingerVector[20], fingerVector[33], fingerVector[34], fingerVector[35]))
		# 16 Distancia de Euclides entre el pulgar y el indice
		# 16 Eunclidean distance betwwen the thumb and the index
		finalVector.append(euclideanDistance(fingerVector[21], fingerVector[22], fingerVector[23], fingerVector[24], fingerVector[25], fingerVector[26]))
		# 17 Distancia de Euclides entre el indice y el corazon
		# 17 Euclidean distance between the index and the middle finger
		finalVector.append(euclideanDistance(fingerVector[24], fingerVector[25], fingerVector[26], fingerVector[27], fingerVector[28], fingerVector[29]))
		# 18 Distancia de Euclides entre el corazon y el anular
		# 18 Euclidean distance between the middle and the ring finger
		finalVector.append(euclideanDistance(fingerVector[27], fingerVector[28], fingerVector[29], fingerVector[30], fingerVector[31], fingerVector[32]))
		# 19 Distancia de Euclides entre el anular y el menique
		# 19 Euclidean distance between the ring and the pinky finger
		finalVector.append(euclideanDistance(fingerVector[30], fingerVector[31], fingerVector[32], fingerVector[33], fingerVector[34], fingerVector[35]))
		# 20 Distancia de Euclides entre el pulgar y el menique
		# 20 Euclidean distance between the thumb and the pinky
		finalVector.append(euclideanDistance(fingerVector[21], fingerVector[22], fingerVector[23], fingerVector[33], fingerVector[34], fingerVector[35]))
		
		# 21 Distancia de euclides entre palmas
		# 21 Euclidean distance between palms
		finalVector.append(euclideanDistance(fingerVector[0], fingerVector[1], fingerVector[2], fingerVector[18], fingerVector[19], fingerVector[20]))

		return finalVector
		
		
	def cosine_similarity(vectorFrame, leftHandNormal, rightHandNormal):
		global dataSet
		cos = 0

		for key in dataSet:
			if len(dataSet[key][0]) == len(vectorFrame):
				vectorNp = numpy.fromiter(vectorFrame, dtype = numpy.int)
				dot = numpy.dot(vectorNp, dataSet[key][0])
				normV1 = numpy.linalg.norm(vectorNp)
				normV2 = numpy.linalg.norm(dataSet[key][0])
				cos = dot / (normV1 * normV2)

				#if (cos > 0.993 and abs(leftHandNormal - float(dataSet[key][1][0])) < 1.0 and abs(rightHandNormal - float(dataSet[key][1][1])) < 1.0): # If you want to change the accuracy of sign recognition you should change the 0.88 number
				if (cos > 0.88 and abs(leftHandNormal - float(dataSet[key][1][0])) < 1.0 and abs(rightHandNormal - float(dataSet[key][1][1])) < 1.0):

					f = open('signRecognisedDataSet.txt','a+')

					if len(vectorFrame) == 10:
						vectorFrame = vectorFrame + [leftHandNormal * 100, rightHandNormal * 100]
					vectorDataBase = str(vectorFrame)
					delete = " []"
					for char in delete:
						vectorDataBase = vectorDataBase.replace(char, "")

					f.write(vectorDataBase + ',' + key[:-1] + '\n')

					print(vectorDataBase + ',' + key[:-1])
					sys.stdout.flush()

					f.close()
					break
		return cos
		
	def normalization(vector): # Este metodo normaliza vectores (no se utiliza pero esta por si acaso)
							   # This methid nomalizes vectors (is not used, but is written just in case)
		bectoh = []
		valor = 0
		for i in vector:
			valor = valor + i ** 2
		
		valor = math.sqrt(valor)
		for i in vector:
			bectoh.append(i/valor)
		return bectoh
		
#------------------------------------ COSINE SIMILARITY ALGORITHM  END ------------------------------------

#------------------------------------ DATASET TO DICTIONARY  BEGIN ------------------------------------

	def csvTest(): 	# Lee un archivo CSV y devuelve un diccionario que tiene como clave el primer valor de cada fila y como 
						# valor un array de numeros que son los demas valores de la fila (en nuestros caso seran las posiciones calculadas)
				   	# Reads a CSV file and returns a dictionary that has as key the first value of the row
				   		# and as value an float arrays which are the rest of the values of the row (in our case are the euclidean distances calculated previously)
		global dataSet	# Inicializamos un diccionario donde se guardaran 
						# Inicialization of the dictionary

		with open('oneHandSignDataSet.txt', 'rb') as csvfile: # se puede pasar un archivo .txt o .csv (pero hay que seguir la estructura)
																				 # it can be a .txt or .csv file (CAREFULL!! no blanks between numbers)
			spamreader = csv.reader(csvfile, delimiter=' ', quotechar='|')
			for row in spamreader:			# bucle for para leer las lineas
											# for loop to read the rows
				palmaNormal = []
				newRow = row[0].split(',')	# Separamos la linea por comas para separa los valores
											# We split the rows by comas in order to get the values
				key = newRow.pop(0)	# Las keys seran el nombre del gesto (es el primer valor de cada linea)
									# The keys will be the name of the sign (first value of every row)

				palmaNormal.append(newRow.pop(-2)) # Estos valores son la orientacion de las manos ( primero la izquierda y despues la derecha) (Mirar documentacion para mas informacion)
				palmaNormal.append(newRow.pop(-1))
												   # These values are the orientation of the hands (first left and then right) (See documentation for more information)

				dataSet[key] = [list(map(float, newRow)),palmaNormal]	# Los valores seran arrays de 10 o 21 (dependiendo del numero de manos) floats que seran las distancias de 
																			# euclides de los dedos de cada gesto
																		# The values of the dictionary will be arrays of 10 or 21 (depending the number of hands) floats which are the
																			# euclidean distances between fingers

			return dataSet	# se devuelve el diccionario creado
		
#------------------------------------ DATASET TO DICTIONARY  END ------------------------------------
		
#------------------------------------ LEAP MOTION CONTROLLER  BEGIN ------------------------------------
	def on_init(self, controller):
		#print ("Initialized!")
		return

	def on_connect(self, controller):
		#print ("Motion Sensor Connected!")
		
		controller.enable_gesture(Leap.Gesture.TYPE_CIRCLE);		# Esto es para habilitar los gestos importados arriba NOTA!: EL PUNTO Y COMA DEL FINAL ES NECESARIO
		controller.enable_gesture(Leap.Gesture.TYPE_KEY_TAP);		# This is to enable the gestures imported above NOTE: THE SEMOCOLON IS NECESSARY
		controller.enable_gesture(Leap.Gesture.TYPE_SCREEN_TAP);	
		controller.enable_gesture(Leap.Gesture.TYPE_SWIPE);			

	def on_disconnect(self, controller):
		#print ("Motion Sensor Disconnected!")
		return


	def on_exit(self, controller):
		#print ("Exited")
		return

	def on_frame(self, controller):
									# Get the most recent frame and report some basic information
		frame = controller.frame()
		
		fingerVector = []
		fingerVectorLeft = [] 	# Esta lista va a contener 6 vectores (18 valores) los cuales son las posiciones (X,Y,Z) de la palma, pulgar, indice, corazon, anular y menique 
									# respectivamente de la mano izquierda
								# This array will contain 6 vectors (18 values), which are the positions (X, Y, Z) of the palm, thumb, index, middle, ring, pinky, respectively 
									# of the left hand
		fingerVectorRight = []  # Esta lista va a contener 6 vectores (18 valores) los cuales son las posiciones (X,Y,Z) de la palma, pulgar, indice, corazon, anular y menique 
									# respectivamente de la mano derecha
								# This array will contain 6 vectors (18 values), which are the positions (X, Y, Z) of the palm, thumb, index, middle, ring, pinky, respectively 
									# of the left hand
		
		leftHandNormal = 0
		rightHandNormal = 0

		# Get hands
		for hand in frame.hands:
			
			handType = "Left hand" if hand.is_left else "Right hand"

			if (handType == 'Left hand'):
				leftHandNormal = hand.palm_normal.y
			else:
				rightHandNormal = hand.palm_normal.y
			
			# NOTA IMPORTANTE!!!: hand.palm_position devuelve un vector the tres floats correspondientes a los valores de los 
				# ejes X, Y, Z respectivamente. Este vector es lo que vamos a aniadir a nuestra lista 
			# IMPORTANT!!!: hand.palm_position return a three float vector corresponding to he valuesof the X, Y, Z axis respectively. This vector will be added to our array

			# Get the hand's normal vector and direction
			normal = hand.palm_normal
			direction = hand.direction

			# NOTA: SE PUEDEN OBTENER MAS DATOS COMO ANGULOS O POSICIONES DEL BRAZO O MUNIECA PERO NO LO NECESITAMOS POR AHORA
			# COMMENT: MORE DATA CAN BE OBTAINED, LIKE ANGLES OR ARM/WRIST POSITION BUT IS NOT NEEDED YET
			
			# Aniadimos la posicion de la palma a la lista fingerVector
			# We append the position of the palm to the fingerVector array
			fingerVector.append(hand.palm_position.x)
			fingerVector.append(hand.palm_position.y)
			fingerVector.append(hand.palm_position.z)
			
			#  Get fingers
			for finger in hand.fingers:
								
				# Get bones
				for b in range(0,4):
					bone = finger.bone(b)
					
					# Aniadimos la posicion donde acaba el hueso "distal" de cada dedo
					# We append the position of the tips of each finger ("distal" bone)
					if b == 3:
						fingerVector.append(bone.next_joint.x)
						fingerVector.append(bone.next_joint.y)
						fingerVector.append(bone.next_joint.z)
				
			if handType == "Left hand":					# Realizamos este if para tener ordenado siempre el vector ( la mano izquierda antes que la mano derecha )
														# We use this if to sort the vector (left hand before right hand)
				fingerVectorLeft = list(fingerVector)	# se utiliza list() para para hacer una clonacion y que al borrar fingerVector no se pierda fingerVectorLeft
														# list() command is used to clone y once we delete fingerVector we dont delete fingerVectorLeft
				del fingerVector[:]
			else:
				fingerVectorRight = list(fingerVector)
				del fingerVector[:]
		
		fingerVector = list(fingerVectorLeft + fingerVectorRight)	# Concatenamos ambos vectores izquierda y derecha con list() 
																	# We concatenate both left and right hand vectors with list()
		
		del fingerVectorLeft[:]
		del fingerVectorRight[:]
				
		
		if fingerVector == []:	# En caso de que no haya ninguna mano y el vector no tenga datos no se haran los calculos
								# If there were no hands detected and the array will not have data, no calculations will be done
			pass
		
		elif len(fingerVector) == 18:							# En caso de que solo se detecte una mano fingerVector sera de 18 ( X Y Z de los dedos y la palma)
																# If only one hand is detected vectorFinger will have 18 values ( X Y Z of the fingers and the palm)
			elVector = vectorCalculated(fingerVector)

			cosine_similarity(elVector, leftHandNormal, rightHandNormal)
			
			del fingerVector[:]
			
		else:															# Sino estara detectando las dos manos es decir un vector con 36 valores
																		# If not two hands will be detected, an array with 36 values
			elVector = vectorCalculated2(fingerVector)
			
			cosine_similarity(elVector, leftHandNormal, rightHandNormal)
			
			del fingerVector[:]
			
#------------------------------------ LEAP MOTION CONTROLLER  END ------------------------------------

#------------------------------------ MAIN BEGIN ------------------------------------
		
def main():
	global control
	listener = LeapMotionListener()
	controller = Leap.Controller()
	csvTest()
	
	controller.add_listener(listener)
			
	#print ("Press enter to quit")
	try:
		sys.stdin.readline()
	except KeyboardInterrupt:
		pass
	finally:
		controller.remove_listener(listener)
		
if __name__ == "__main__":
	main()
	
#------------------------------------ MAIN END ------------------------------------